<?php

namespace Drupal\username_phone\Form;

use Drupal\Core\Entity\EntityConstraintViolationListInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\user\RegisterForm;

/**
 * Form handler for the user register forms.
 *
 * @internal
 */
class UsernamePhoneForm extends RegisterForm {

  /**
   * {@inheritdoc}
   */
  public function form(array $form, FormStateInterface $form_state) {
    // Start with the default user account fields.
    $form = parent::form($form, $form_state);

    /** @var \Drupal\user\UserInterface $account */
    $account = $this->entity;

    // Check for new account.
    $register = $account->isNew();

    // Load Username phone settings.
    $config = \Drupal::config('username_phone.settings');
    $configUsername = \Drupal::config('username.settings');
    $username = $configUsername->get('username');

    // Add phone field in user account form.
    $form['account']['phone'] = [
      '#type' => 'phone',
      '#title' => t('Phone number'),
      '#description' => t('The phone number is not made public. It will only be used if you need to be contacted about your account or for opted-in notifications.'),
      '#required' => (isset($username['phone']) || $configUsername->get('phone.verification')) && !(!$account->getPhone() && \Drupal::currentUser()->hasPermission('administer users')),
      '#default_value' => (!$register ? $account->getPhone() : ''),
      '#access' => $account->phone->access('edit'),
      '#countries' => $config->get('countries'),
      '#exclude_countries' => [],
      '#initial_country' => $config->get('initial_country'),
      '#preferred_countries' => $config->get('preferred_countries'),
      '#show_flags' => $config->get('show_flags'),
      '#allow_dropdown' => $config->get('allow_dropdown'),
      '#auto_insert_dial_code' => $config->get('auto_insert_dial_code'),
      '#separate_dial_code' => $config->get('separate_dial_code'),
    ];
    $form['account']['phone_number'] = [
      '#type' => 'hidden',
      '#attributes' => ['class' => 'temporary-phone']
    ];
    $form['account']['country_code'] = [
      '#type' => 'hidden',
      '#attributes' => ['class' => 'temporary-phone']
    ];
    $form['account']['country_iso2'] = [
      '#type' => 'hidden',
      '#attributes' => ['class' => 'temporary-phone']
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  protected function actions(array $form, FormStateInterface $form_state) {
    return parent::actions($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    $values = $form_state->getValues();
    if (isset($values['phone_number'])) {
      $phone_mask = $values['phone'];
      $form_state->setValue('phone', $values['phone_number']);
      $form_state->setValue('phone_number', $phone_mask);
    }

    parent::validateForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    parent::submitForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function save(array $form, FormStateInterface $form_state) {
    parent::save($form, $form_state);

    $account = $this->entity;
    $values = $form_state->getValues();
    if (isset($values['phone']) && isset($values['phone_number']) && isset($values['country_code']) && isset($values['country_iso2'])) {
      // Save to the database and the static cache.
      \Drupal::database()->merge('user__phones')
        ->key(['uid' => $account->id()])
        ->fields([
          'phone_number' => $values['phone'],
          'phone_mask' => $values['phone_number'],
          'country_code' => $values['country_code'],
          'country_iso2' => $values['country_iso2'],
          'created' => \Drupal::time()->getRequestTime(),
        ])
        ->execute();
    }
  }

  /**
   * {@inheritdoc}
   */
  protected function flagViolations(EntityConstraintViolationListInterface $violations, array $form, FormStateInterface $form_state) {
    // Manually add phone flag violations of user phone field not handled by
    // the form display. This is necessary as entity form displays only flag
    // violations for fields contained in the display.
    $field_names = [
      'phone',
    ];
    foreach ($violations->getByFields($field_names) as $violation) {
      [$field_name] = explode('.', $violation->getPropertyPath(), 2);
      $form_state->setErrorByName($field_name, $violation->getMessage());
    }
    parent::flagViolations($violations, $form, $form_state);
  }

}
